# --------------------------------------------------------------------
# --   *****************************
# --   *   Trenz Electronic GmbH   *
# --   *   Beendorfer Straße 23    *
# --   *   32609 Hüllhorst         *
# --   *   Germany                 *
# --   *****************************
# --------------------------------------------------------------------
# -- $Author: Dück, Thomas $
# -- $Email: t.dueck@trenz-electronic.de $
# --------------------------------------------------------------------
# -- Change History:
# ------------------------------------------
# -- $Date: 2024/06/06 | $Author: Dück, Thomas
# -- - initial release
# --------------------------------------------------------------------
# --------------------------------------------------------------------
namespace eval ::TE {
 namespace eval EXT {
  #--------------------------------
  #-- run_serial: 
  proc run_serial {{serial NA}  {metadata 0} {factoryorder 0}} {
    set val "Scanning not available"
    set command exec
    lappend command powershell${::TE::WIN_EXE}
    lappend command -file  
    lappend command ${::TE::SERIAL_PATH}/getartikelbyserial.ps1
    lappend command $serial
    lappend command ${::TE::SERIAL_PATH}
    lappend command ${metadata}
    lappend command ${factoryorder}
    lappend command "0"

    if {[catch {set val [eval $command]} result]} {set val "$result"}

    return $val
  }

  #--------------------------------
  #-- run_putty: under development, only for WinOS
  proc run_putty {{filename ""}} {
    if {$::tcl_platform(platform) eq "windows"} {
      # putty -serial com12 -sercfg 115200,8,n,1,N
      set command exec
      if { [file exists ${::TE::COM_PATH}/putty.exe] } {
        lappend command ${::TE::COM_PATH}/putty.exe
      } else {
        lappend command putty
      }
      lappend command -serial ${::TE::DESIGN_UART_COM} 
      lappend command -sercfg ${::TE::DESIGN_UART_SPEED},8,n,1,N
      if {[llength $filename] > 4} {
        lappend command -sessionlog ${::TE::LOG_PATH}/$filename
      } elseif {$filename eq "d"} {
      } else {
        lappend command -sessionlog ${::TE::LOG_PATH}/putty-${::TE::PRODID}-&y&m&d-&t.log
      }
      lappend command &
      set pid [eval $command]
      return $pid
    } else {
      return "Not supported on this OS"
    }
  }

  #--------------------------------
  #-- find_com_port: search com port for serial console, only for WinOS
  proc find_com_port {} {  
    set command exec
    lappend command powershell${::TE::WIN_EXE}
    lappend command -command
    # lappend command "Get-PnpDevice -PresentOnly | Where-Object { \$_.InstanceId -like \"*VID_0403*PID_6010*\" } | Select-Object Status,Class,FriendlyName,InstanceId | Format-List"
    lappend command "Get-PnpDevice -Class Ports -PresentOnly | Where-Object { \$_.InstanceId -like \"FTDIBUS*\" } | Select-Object Status,Class,FriendlyName,InstanceId | Format-List"
    set result ""
    [catch {set val [eval $command]} result]
    ::TE::UTILS::report -msg $result -command $command -msgid TE_EXT-01 -silent
    set result [string map {" " ""} $result]
    
    set comportlist [list]
    foreach line $result {
      if { $line ne "" && [string match "*COM*" $line]} {
        regexp -nocase -line {COM(\w+)} $line comport
        if {[lsearch $::TE::COM_IGNORE_LIST $comport] eq -1} {
          lappend comportlist $comport
        }
      } 
    }
    
    if {$comportlist eq ""} { set comportlist "---" }
    
    return $comportlist
  }
  
  #--------------------------------
  #-- zip_project: 
  proc zip_project { EXCLUDELIST ZIPNAME {release "NA"} {initials "NA"} {dest "NA"} {ttyp "NA"} {btyp "NA"} {pext "NA"}} {

    if {$dest eq "NA"} {
      set someVar "NA"
      puts "---------------------------------------------"
      puts "--Additional Information for ZIP are required"
      puts "---------------------------------------------"
      puts "Insered Initials(optional):"
      while {1} {
        gets stdin someVar
        if {![string match -nocase "" $someVar ]} {
          set initials $someVar
          break
        } 
      }
      
      puts "Insered destination:"
      puts "- 0 for PublicDoc"
      puts "- 1 for Production"
      puts "- 2 for Development"
      puts "- 3 for Preliminary"

      while {1} {
        set someVar "NA"
        gets stdin someVar
        if {[string match -nocase "0" $someVar ]} {
          set dest "PublicDoc"
          break
        } elseif {[string match -nocase "1" $someVar ]} {
          set dest "Production"
          set someVar "NA"
          while {1} {
            puts "Insered destination:"
            puts "- 0 for Manual Test"
            puts "- 1 for Halfautomatic Test"
            puts "- 2 for Automatic Test System"
            puts "- 3 for Others"
            gets stdin someVar
            if {[string match -nocase "0" $someVar ]} {
              set ttyp "MT"
              break
            } elseif {[string match -nocase "1" $someVar ]} {
              set ttyp "HT"
              break
            } elseif {[string match -nocase "2" $someVar ]} {
              set ttyp "ATS"
              break
            } elseif {[string match -nocase "3" $someVar ]} {
              set ttyp "NA"
              break
            }
          }
          set someVar "NA"
          while {1} {
            puts "Insered Board Type(Test Reason):"
            puts "- 0 for Module Test Export"
            puts "- 1 for Carrier Test Export"
            puts "- 2 for Motherboard Test Export"
            puts "- 3 for FMC-Card Test Export"
            puts "- 4 for PCIe-Card Test Export"
            puts "- 5 for Others"
            gets stdin someVar
            if {[string match -nocase "0" $someVar ]} {
              set btyp "Module"
              break
            } elseif {[string match -nocase "1" $someVar ]} {
              set btyp "Carrier"
              break
            } elseif {[string match -nocase "2" $someVar ]} {
              set btyp "Motherboard"
              break
            } elseif {[string match -nocase "3" $someVar ]} {
              set btyp "FMC-Card"
              break
            } elseif {[string match -nocase "4" $someVar ]} {
              set btyp "PCIe-Card"
              break
            } elseif {[string match -nocase "5" $someVar ]} {
              set btyp "NA"
              break
            }
          }
          set someVar "NA"
          while {1} {
              puts "Include init.sh extention:"
              puts "- 0 with extentions"
              puts "- 1 without extentions"
              gets stdin someVar
            if {[string match -nocase "0" $someVar ]} {
              set pext "yes"
              break
            } elseif {[string match -nocase "1" $someVar ]} {
              set pext "NA"
              break
            }
          }
          break
        } elseif {[string match -nocase "2" $someVar ]} {
          set dest "Development"
          break
        } elseif {[string match -nocase "3" $someVar ]} {
          set dest "Preliminary"
          break
        }
      }
    }
    
    ::TE::UTILS::te_msg -type info -id TE_EXT-02 -msg "Zip Project. Please wait ..."  
    
    # source path
    set sourcepath [string trim ${::TE::QPROJ_PATH} "quartus"]
    set destinationpath ${::TE::BACKUP_PATH}/${::TE::QPROJ_SRC_NAME}
    set zipteinfo_path "${destinationpath}/settings"
    
    #remove old backup project copy
    if {[file exists ${destinationpath}]} { 
      file delete -force ${destinationpath}  
    }
    #create new destination folder
    file mkdir ${destinationpath}
    
    #get all files
    set filelist [ glob ${sourcepath}*]
    #remove backup folder
    set findex [lsearch $filelist *backup]
    set filelist [lreplace $filelist[set filelist {}] $findex $findex]
    #remove console folder if dest==PublicDoc/Preliminary
    if {[string match "PublicDoc" ${dest}] || [string match "Preliminary" ${dest}]} { 
      set findex [lsearch $filelist *console]
      set filelist [lreplace $filelist[set filelist {}] $findex $findex] 
    }
      
    foreach el $filelist {
      file copy -force ${el} ${destinationpath}
    }

    foreach el $EXCLUDELIST {
      set find ""
      if {[catch {set find [glob -join -dir $destinationpath $el]}] && [catch {set find [glob -type hidden -join -dir $destinationpath $el]}]} {
        ::TE::UTILS::te_msg -type info -id TE_EXT-03 -msg "$el doesn't exist."
      } else {
        ::TE::UTILS::te_msg -type info -id TE_EXT-04 -msg "Excluded from backup: $find"
        file delete -force $find
      }
    }
    # remove .svn folder from reference design backup
    ::TE::UTILS::delete_files -n ".svn" -t "d" -d $destinationpath -s
    
    # modify design_basic_settings.tcl for production
    if {[string match "Production" ${dest}]} {
      set fp_r [open "${destinationpath}/settings/design_basic_settings.tcl" r]
      set file_data [read $fp_r]
      close $fp_r
      set data [split $file_data "\n"]
      set fp_w [open "${destinationpath}/settings/design_basic_settings.tcl" w]
      foreach line $data {
        if { [string match "QUARTUS_PATH_WIN=*" $line] }   { set line "QUARTUS_PATH_WIN=C:/intelFPGA_pro" }
        if { [string match "QUARTUS_PATH_LINUX=*" $line] } { set line "QUARTUS_PATH_LINUX=~/intelFPGA_pro" }
        if { [string match "QUARTUS_VERSION=*" $line] }    { set line "QUARTUS_VERSION=24.1" }
        if { [string match "QUARTUS_EDITION=*" $line] }    { set line "QUARTUS_EDITION=Pro" }
        if { [string match "QUARTUS_PROG=*" $line] }       { set line "QUARTUS_PROG=1" }
        puts $fp_w "$line"
      }
      close $fp_w
    }
    #add _dev for dest==Development and _prelim for dest==Preliminary
    if {[string match "Development" ${dest}]} {
      set ZIPNAME "$ZIPNAME_dev"
    } elseif {[string match "Preliminary" ${dest}]} {
      set ZIPNAME "$ZIPNAME_prelim"
    }
    # write zipinfo and zip project
    ::TE::EXP::write_zip_info $zipteinfo_path $ZIPNAME $release $initials $dest $ttyp $btyp "NA"
    ::TE::EXT::zip_general ${destinationpath} ${ZIPNAME}

    # in case with production extentions:
    if {[string match "yes" ${pext}]} {
      #remove project copy
      if {[file exists ${destinationpath}]} { 
        file delete -force ${destinationpath}  
      }
      set destinationpath "${TE::BACKUP_PATH}/Produktionstest"
      if {[file exists ${destinationpath}]} { 
        file delete -force ${destinationpath}  
      }
      file mkdir ${destinationpath}
      file copy -force ${TE::BACKUP_PATH}/${ZIPNAME}.zip ${destinationpath}
      file delete -force ${TE::BACKUP_PATH}/${ZIPNAME}.zip 
      file copy -force "${TE::BASEFOLDER}/../prod_cfg_list.csv" ${destinationpath}
      file copy -force "${TE::BASEFOLDER}/../cfg_init" ${destinationpath}
      # remove .svn folder from production test zip
      ::TE::UTILS::delete_files -n ".svn" -t "d" -d ${destinationpath} -s
        
      set ZIPNAME "${ZIPNAME}_extended"
      zip_general "${::TE::BACKUP_PATH}/Produktionstest/" $ZIPNAME
    }

    #remove project copy
    if {[file exists ${destinationpath}]} { 
      file delete -force ${destinationpath}
    }

    ::TE::UTILS::te_msg -type info -id TE_EXT-05 -msg "Zip Project -> done"
    ::TE::UTILS::te_msg -msg "------------------------------"
  }
  
  #--------------------------------
  #-- zip_general: zip project
  proc zip_general { ZIPPATH ZIPNAME } {
    set command exec
    if {[file exists $::quartus(binpath)7z${::TE::WIN_EXE}]} {
      puts "7z: $ZIPPATH"
      lappend command 7z${::TE::WIN_EXE}
      lappend command a
      lappend command -tzip 
      lappend command -o${::TE::BACKUP_PATH}
      #lappend command -r  
      lappend command "${ZIPNAME}.zip"
      lappend command ${ZIPPATH}  
    } else {
      lappend command zip${::TE::WIN_EXE}
      #lappend command -r
      lappend command ${ZIPNAME}.zip
      lappend command ${ZIPPATH}
    }

    cd ${::TE::BACKUP_PATH}
    catch {eval $command} result
    ::TE::UTILS::te_msg -type info -id TE_EXT-06 -msg "Result of command $command:\n$result" 
    cd ${::TE::BASEFOLDER}  
  }

 }
 
  # -----------------------------------------------------------------------------------------------------------------------------------------
  # user command functions
  # -----------------------------------------------------------------------------------------------------------------------------------------
  #--------------------------------
  #-- pr_putty: Requires Putty installation
  proc pr_putty {{args ""}} {
    set DISPLAY_COM false
    set LOGNAME ""
    set predef_comport ${::TE::DESIGN_UART_COM}

    set args_cnt [llength $args]
    for {set option 0} {$option < $args_cnt} {incr option} {
      switch [lindex $args $option] { 
        "-available_com"  { set DISPLAY_COM             true }
        "-com"            { set ::TE::DESIGN_UART_COM   [lindex $args $option+1];  incr option; }
        "-speed"          { set ::TE::DESIGN_UART_SPEED [lindex $args $option+1];  incr option; }
        "-log"            { set LOGNAME                 [lindex $args $option+1];  incr option; }
        ""                {}
        default           { ::TE::UTILS::te_msg -type error -id TE_PR-01 -msg "Unrecognised argument: pr_putty [lindex $args $option]."
                            ::TE::UTILS::te_msg -msg "Expected arguments: pr_putty \[options\]"
                            ::TE::UTILS::te_msg -msg "     Options:"
                            ::TE::UTILS::te_msg -msg "          -available_com"
                            ::TE::UTILS::te_msg -msg "          -com   \"<com port>\""
                            ::TE::UTILS::te_msg -msg "          -speed \"<baudrate>\""
                            ::TE::UTILS::te_msg -msg "          -log   \"<log file name>\""
                            return
                          }
      }
    }
    
    ::TE::UTILS::te_msg -type info -id TE_PR-02 -msg "Start Putty console"
    if { ${DISPLAY_COM} } {
      ::TE::EXT::find_com_port
    } else {
      if { ${::TE::DESIGN_UART_COM} eq "NA" } {
        ::TE::UTILS::te_msg -type warning -id TE_PR-03 -msg "Scripts try auto detect of COM port, if failed, try to open putty and connect to COM port manually."
      }        
      if { ${::TE::DESIGN_UART_SPEED} eq "NA"} {
        ::TE::UTILS::te_msg -type error -id TE_PR-04
      }
      if { [catch {   
        set tmpcomlist [::TE::EXT::find_com_port]
        if { $tmpcomlist eq "" } { 
          ::TE::UTILS::te_msg -type error -id TE_PR-05 -msg "No COM ports found."
          return
        } else {
          ::TE::UTILS::te_msg -msg "Available COM Ports: $tmpcomlist"
        }
        
        foreach line $tmpcomlist {
          if {[string match "${::TE::DESIGN_UART_COM}" $line]} { set tmpcomlist ${::TE::DESIGN_UART_COM} }
        }
        
        foreach cur_comport $tmpcomlist {
          set ::TE::DESIGN_UART_COM $cur_comport
          ::TE::UTILS::te_msg -type info -id TE_PR-06 -msg "Open Putty on ${::TE::DESIGN_UART_COM} with speed ${::TE::DESIGN_UART_SPEED}"
          set printout [::TE::EXT::run_putty $LOGNAME]
          ::TE::UTILS::te_msg -type info -id TE_PR-07 -msg "Putty is opened with PID: $printout"
        }
      } result] } {::TE::UTILS::te_msg -type error -id TE_PR-08 -msg "Script (::TE::EXT::run_putty) failed: $result."; return -code error}
    }

    set ::TE::DESIGN_UART_COM $predef_comport
    ::TE::UTILS::te_msg -type info -id TE_PR-09 -msg "Start Putty console finished."
  }

  # -----------------------------------------------------------------------------------------------------------------------------------------
  # finished user command functions
  # -----------------------------------------------------------------------------------------------------------------------------------------


  ::TE::UTILS::te_msg -type info -msg "(TE) Load External Script finished"
}